@@ -18,14 +18,15 @@ const rootdirs = [
     "../contrib",
     "../src",
     "../stdlib",
-    "../test",
 ]
 
 # to exculde whole sub directories
 const excludedirs = [
     # see: https://github.com/JuliaLang/julia/pull/11073#issuecomment-98090053
-    "../base/grisu",
+    "../base/ryu",
     "../src/flisp",
+    "../stdlib/TOML/test/testfiles",
+    "../test/testhelpers/allocation_file.jl",
 ]
 
 const skipfiles = [
