@@ -31,6 +31,14 @@ static int is_bom(uint32_t wc)
     return wc == 0xFEFF;
 }
 
+static int safe_peekutf8(fl_context_t *fl_ctx, ios_t *s, uint32_t *pwc)
+{
+    int result = ios_peekutf8(s, pwc);
+    if (result == 0)
+        lerror(fl_ctx, fl_ctx->IOError, "invalid UTF-8 sequence");
+    return result;
+}
+
 value_t fl_skipws(fl_context_t *fl_ctx, value_t *args, uint32_t nargs)
 {
     argcount(fl_ctx, "skip-ws", nargs, 2);
@@ -39,7 +47,7 @@ value_t fl_skipws(fl_context_t *fl_ctx, value_t *args, uint32_t nargs)
     uint32_t wc=0;
     value_t skipped = fl_ctx->F;
     while (1) {
-        if (ios_peekutf8(s, &wc) == IOS_EOF) {
+        if (safe_peekutf8(fl_ctx, s, &wc) == IOS_EOF) {
             ios_getutf8(s, &wc);  // to set EOF flag if this is a true EOF
             if (!ios_eof(s))
                 lerror(fl_ctx, symbol(fl_ctx, "error"), "incomplete character");
@@ -62,8 +70,11 @@ static int is_wc_cat_id_start(uint32_t wc, utf8proc_category_t cat)
             cat == UTF8PROC_CATEGORY_LT || cat == UTF8PROC_CATEGORY_LM ||
             cat == UTF8PROC_CATEGORY_LO || cat == UTF8PROC_CATEGORY_NL ||
             cat == UTF8PROC_CATEGORY_SC ||  // allow currency symbols
-            // other symbols, but not arrows
-            (cat == UTF8PROC_CATEGORY_SO && !(wc >= 0x2190 && wc <= 0x21FF)) ||
+            // other symbols, but not arrows or replacement characters
+            (cat == UTF8PROC_CATEGORY_SO && !(wc >= 0x2190 && wc <= 0x21FF) &&
+             wc != 0xfffc && wc != 0xfffd &&
+             wc != 0x233f &&  // notslash
+             wc != 0x00a6) || // broken bar
 
             // math symbol (category Sm) whitelist
             (wc >= 0x2140 && wc <= 0x2a1c &&
@@ -106,7 +117,10 @@ static int is_wc_cat_id_start(uint32_t wc, utf8proc_category_t cat)
 
             // Other_ID_Start
             wc == 0x2118 || wc == 0x212E || // ℘, ℮
-            (wc >= 0x309B && wc <= 0x309C)); // katakana-hiragana sound marks
+            (wc >= 0x309B && wc <= 0x309C) || // katakana-hiragana sound marks
+
+            // bold-digits and double-struck digits
+            (wc >= 0x1D7CE && wc <= 0x1D7E1)); // 𝟎 through 𝟗 (inclusive), 𝟘 through 𝟡 (inclusive)
 }
 
 JL_DLLEXPORT int jl_id_start_char(uint32_t wc)
@@ -140,9 +154,9 @@ JL_DLLEXPORT int jl_id_char(uint32_t wc)
 #include "julia_opsuffs.h"
 
 // chars that can follow an operator (e.g. +) and be parsed as part of the operator
-int jl_op_suffix_char(uint32_t wc)
+JL_DLLEXPORT int jl_op_suffix_char(uint32_t wc)
 {
-    static htable_t jl_opsuffs;
+    static htable_t jl_opsuffs; // XXX: requires uv_once
     if (!jl_opsuffs.size) { // initialize hash table of suffixes
         size_t i, opsuffs_len = sizeof(opsuffs) / (sizeof(uint32_t));
         htable_t *h = htable_new(&jl_opsuffs, opsuffs_len);
@@ -238,6 +252,7 @@ value_t fl_julia_strip_op_suffix(fl_context_t *fl_ctx, value_t *args, uint32_t n
     if (!op[i]) return args[0]; // no suffix to strip
     if (!i) return args[0]; // only suffix chars --- might still be a valid identifier
     char *opnew = strncpy((char*)malloc(i+1), op, i);
+    // TODO: if argument to opnew == NULL
     opnew[i] = 0;
     value_t opnew_symbol = symbol(fl_ctx, opnew);
     free(opnew);
@@ -259,7 +274,7 @@ value_t fl_julia_underscore_symbolp(fl_context_t *fl_ctx, value_t *args, uint32_
 
 utf8proc_int32_t jl_charmap_map(utf8proc_int32_t c, void *ctx)
 {
-    static htable_t jl_charmap;
+    static htable_t jl_charmap; // XXX: requires uv_once
     if (!jl_charmap.size) { // initialize hash table
         size_t i, charmap_len = sizeof(charmap) / (2*sizeof(uint32_t));
         htable_t *h = htable_new(&jl_charmap, charmap_len);
@@ -314,13 +329,13 @@ value_t fl_accum_julia_symbol(fl_context_t *fl_ctx, value_t *args, uint32_t narg
         type_error(fl_ctx, "accum-julia-symbol", "wchar", args[0]);
     uint32_t wc = *(uint32_t*)cp_data((cprim_t*)ptr(args[0]));
     ios_t str;
-    int allascii=1;
+    int allascii = 1;
     ios_mem(&str, 0);
     do {
         allascii &= (wc <= 0x7f);
         ios_getutf8(s, &wc);
         if (wc == '!') {
-            uint32_t nwc;
+            uint32_t nwc = 0;
             ios_peekutf8(s, &nwc);
             // make sure != is always an operator
             if (nwc == '=') {
@@ -329,7 +344,7 @@ value_t fl_accum_julia_symbol(fl_context_t *fl_ctx, value_t *args, uint32_t narg
             }
         }
         ios_pututf8(&str, wc);
-        if (ios_peekutf8(s, &wc) == IOS_EOF)
+        if (safe_peekutf8(fl_ctx, s, &wc) == IOS_EOF)
             break;
     } while (jl_id_char(wc));
     ios_pututf8(&str, 0);
