@@ -6,17 +6,17 @@
   . pool-allocates small objects, keeps big objects on a simple list
 */
 
-#ifndef JULIA_GC_H
-#define JULIA_GC_H
+#ifndef JL_GC_H
+#define JL_GC_H
 
 #include <stdlib.h>
 #include <string.h>
 #ifndef _MSC_VER
 #include <strings.h>
 #endif
-#include <assert.h>
 #include <inttypes.h>
 #include "julia.h"
+#include "julia_threads.h"
 #include "julia_internal.h"
 #include "threading.h"
 #ifndef _OS_WINDOWS_
@@ -25,6 +25,7 @@
 #define MAP_ANONYMOUS MAP_ANON
 #endif
 #endif
+#include "julia_assert.h"
 
 #ifdef __cplusplus
 extern "C" {
@@ -74,9 +75,167 @@ typedef struct {
     int         full_sweep;
 } jl_gc_num_t;
 
+enum {
+    GC_MARK_L_marked_obj,
+    GC_MARK_L_scan_only,
+    GC_MARK_L_finlist,
+    GC_MARK_L_objarray,
+    GC_MARK_L_array8,
+    GC_MARK_L_array16,
+    GC_MARK_L_obj8,
+    GC_MARK_L_obj16,
+    GC_MARK_L_obj32,
+    GC_MARK_L_stack,
+    GC_MARK_L_excstack,
+    GC_MARK_L_module_binding,
+    _GC_MARK_L_MAX
+};
+
+// The following structs (`gc_mark_*_t`) contain iterator state used for the
+// scanning of various object types.
+//
+// The `nptr` member records the number of pointers slots referenced by
+// an object to be used in the full collection heuristics as well as whether the object
+// references young objects.
+// `nptr >> 2` is the number of pointers fields referenced by the object.
+// The lowest bit of `nptr` is set if the object references young object.
+// The 2nd lowest bit of `nptr` is the GC old bits of the object after marking.
+// A `0x3` in the low bits means that the object needs to be in the remset.
+
+// An generic object that's marked and needs to be scanned
+// The metadata might need update too (depend on the PC)
+typedef struct {
+    jl_value_t *obj; // The object
+    uintptr_t tag; // The tag with the GC bits masked out
+    uint8_t bits; // The GC bits after tagging (`bits & 1 == 1`)
+} gc_mark_marked_obj_t;
+
+// An object array. This can come from an array, svec, or the using array or a module
+typedef struct {
+    jl_value_t *parent; // The parent object to trigger write barrier on.
+    jl_value_t **begin; // The first slot to be scanned.
+    jl_value_t **end; // The end address (after the last slot to be scanned)
+    uint32_t step; // Number of pointers to jump between marks
+    uintptr_t nptr; // See notes about `nptr` above.
+} gc_mark_objarray_t;
+
+// A normal object with 8bits field descriptors
+typedef struct {
+    jl_value_t *parent; // The parent object to trigger write barrier on.
+    uint8_t *begin; // Current field descriptor.
+    uint8_t *end; // End of field descriptor.
+    uintptr_t nptr; // See notes about `nptr` above.
+} gc_mark_obj8_t;
+
+// A normal object with 16bits field descriptors
+typedef struct {
+    jl_value_t *parent; // The parent object to trigger write barrier on.
+    uint16_t *begin; // Current field descriptor.
+    uint16_t *end; // End of field descriptor.
+    uintptr_t nptr; // See notes about `nptr` above.
+} gc_mark_obj16_t;
+
+// A normal object with 32bits field descriptors
+typedef struct {
+    jl_value_t *parent; // The parent object to trigger write barrier on.
+    uint32_t *begin; // Current field descriptor.
+    uint32_t *end; // End of field descriptor.
+    uintptr_t nptr; // See notes about `nptr` above.
+} gc_mark_obj32_t;
+
+typedef struct {
+    jl_value_t **begin; // The first slot to be scanned.
+    jl_value_t **end; // The end address (after the last slot to be scanned)
+    uint8_t *rebegin;
+    gc_mark_obj8_t elem;
+} gc_mark_array8_t;
+
+typedef struct {
+    jl_value_t **begin; // The first slot to be scanned.
+    jl_value_t **end; // The end address (after the last slot to be scanned)
+    uint16_t *rebegin;
+    gc_mark_obj16_t elem;
+} gc_mark_array16_t;
+
+// Stack frame
+typedef struct {
+    jl_gcframe_t *s; // The current stack frame
+    uint32_t i; // The current slot index in the frame
+    uint32_t nroots; // `nroots` fields in the frame
+    // Parameters to mark the copy_stack range.
+    uintptr_t offset;
+    uintptr_t lb;
+    uintptr_t ub;
+} gc_mark_stackframe_t;
+
+// Exception stack data
+typedef struct {
+    jl_excstack_t *s;   // Stack of exceptions
+    size_t itr;         // Iterator into exception stack
+    size_t bt_index;    // Current backtrace buffer entry index
+    size_t jlval_index; // Index into GC managed values for current bt entry
+} gc_mark_excstack_t;
+
+// Module bindings. This is also the beginning of module scanning.
+// The loop will start marking other references in a module after the bindings are marked
+typedef struct {
+    jl_module_t *parent; // The parent module to trigger write barrier on.
+    jl_binding_t **begin; // The first slot to be scanned.
+    jl_binding_t **end; // The end address (after the last slot to be scanned)
+    uintptr_t nptr; // See notes about `nptr` above.
+    uint8_t bits; // GC bits of the module (the bits to mark the binding buffer with)
+} gc_mark_binding_t;
+
+// Finalizer (or object) list
+typedef struct {
+    jl_value_t **begin;
+    jl_value_t **end;
+} gc_mark_finlist_t;
+
+// This is used to determine the max size of the data objects on the data stack.
+// We'll use this size to determine the size of the data stack corresponding to a
+// PC stack size. Since the data objects are not all of the same size, we'll waste
+// some memory on the data stack this way but that size is unlikely going to be significant.
+union _jl_gc_mark_data {
+    gc_mark_marked_obj_t marked;
+    gc_mark_objarray_t objarray;
+    gc_mark_array8_t array8;
+    gc_mark_array16_t array16;
+    gc_mark_obj8_t obj8;
+    gc_mark_obj16_t obj16;
+    gc_mark_obj32_t obj32;
+    gc_mark_stackframe_t stackframe;
+    gc_mark_excstack_t excstackframe;
+    gc_mark_binding_t binding;
+    gc_mark_finlist_t finlist;
+};
+
+// Pop a data struct from the mark data stack (i.e. decrease the stack pointer)
+// This should be used after dispatch and therefore the pc stack pointer is already popped from
+// the stack.
+STATIC_INLINE void *gc_pop_markdata_(jl_gc_mark_sp_t *sp, size_t size)
+{
+    jl_gc_mark_data_t *data = (jl_gc_mark_data_t *)(((char*)sp->data) - size);
+    sp->data = data;
+    return data;
+}
+#define gc_pop_markdata(sp, type) ((type*)gc_pop_markdata_(sp, sizeof(type)))
+
+// Re-push a frame to the mark stack (both data and pc)
+// The data and pc are expected to be on the stack (or updated in place) already.
+// Mainly useful to pause the current scanning in order to scan an new object.
+STATIC_INLINE void *gc_repush_markdata_(jl_gc_mark_sp_t *sp, size_t size) JL_NOTSAFEPOINT
+{
+    jl_gc_mark_data_t *data = sp->data;
+    sp->pc++;
+    sp->data = (jl_gc_mark_data_t *)(((char*)sp->data) + size);
+    return data;
+}
+#define gc_repush_markdata(sp, type) ((type*)gc_repush_markdata_(sp, sizeof(type)))
+
 // layout for big (>2k) objects
 
-typedef struct _bigval_t {
+JL_EXTENSION typedef struct _bigval_t {
     struct _bigval_t *next;
     struct _bigval_t **prev; // pointer to the next field of the prev entry
     union {
@@ -212,18 +371,20 @@ typedef struct {
     int ub;
 } pagetable_t;
 
+#ifdef __clang_analyzer__
+unsigned ffs_u32(uint32_t bitvec) JL_NOTSAFEPOINT;
+#else
 STATIC_INLINE unsigned ffs_u32(uint32_t bitvec)
 {
-#if defined(_COMPILER_MINGW_)
-    return __builtin_ffs(bitvec) - 1;
-#elif defined(_COMPILER_MICROSOFT_)
+#if defined(_COMPILER_MICROSOFT_)
     unsigned long j;
     _BitScanForward(&j, bitvec);
     return j;
 #else
-    return ffs(bitvec) - 1;
+    return __builtin_ffs(bitvec) - 1;
 #endif
 }
+#endif
 
 extern jl_gc_num_t gc_num;
 extern pagetable_t memory_map;
@@ -232,55 +393,55 @@ extern arraylist_t finalizer_list_marked;
 extern arraylist_t to_finalize;
 extern int64_t lazy_freed_pages;
 
-STATIC_INLINE bigval_t *bigval_header(jl_taggedvalue_t *o)
+STATIC_INLINE bigval_t *bigval_header(jl_taggedvalue_t *o) JL_NOTSAFEPOINT
 {
     return container_of(o, bigval_t, header);
 }
 
 // round an address inside a gcpage's data to its beginning
-STATIC_INLINE char *gc_page_data(void *x)
+STATIC_INLINE char *gc_page_data(void *x) JL_NOTSAFEPOINT
 {
     return (char*)(((uintptr_t)x >> GC_PAGE_LG2) << GC_PAGE_LG2);
 }
 
-STATIC_INLINE jl_taggedvalue_t *page_pfl_beg(jl_gc_pagemeta_t *p)
+STATIC_INLINE jl_taggedvalue_t *page_pfl_beg(jl_gc_pagemeta_t *p) JL_NOTSAFEPOINT
 {
     return (jl_taggedvalue_t*)(p->data + p->fl_begin_offset);
 }
 
-STATIC_INLINE jl_taggedvalue_t *page_pfl_end(jl_gc_pagemeta_t *p)
+STATIC_INLINE jl_taggedvalue_t *page_pfl_end(jl_gc_pagemeta_t *p) JL_NOTSAFEPOINT
 {
     return (jl_taggedvalue_t*)(p->data + p->fl_end_offset);
 }
 
-STATIC_INLINE int gc_marked(uintptr_t bits)
+STATIC_INLINE int gc_marked(uintptr_t bits) JL_NOTSAFEPOINT
 {
     return (bits & GC_MARKED) != 0;
 }
 
-STATIC_INLINE int gc_old(uintptr_t bits)
+STATIC_INLINE int gc_old(uintptr_t bits) JL_NOTSAFEPOINT
 {
     return (bits & GC_OLD) != 0;
 }
 
-STATIC_INLINE uintptr_t gc_set_bits(uintptr_t tag, int bits)
+STATIC_INLINE uintptr_t gc_set_bits(uintptr_t tag, int bits) JL_NOTSAFEPOINT
 {
     return (tag & ~(uintptr_t)3) | bits;
 }
 
-STATIC_INLINE uintptr_t gc_ptr_tag(void *v, uintptr_t mask)
+STATIC_INLINE uintptr_t gc_ptr_tag(void *v, uintptr_t mask) JL_NOTSAFEPOINT
 {
     return ((uintptr_t)v) & mask;
 }
 
-STATIC_INLINE void *gc_ptr_clear_tag(void *v, uintptr_t mask)
+STATIC_INLINE void *gc_ptr_clear_tag(void *v, uintptr_t mask) JL_NOTSAFEPOINT
 {
     return (void*)(((uintptr_t)v) & ~mask);
 }
 
 NOINLINE uintptr_t gc_get_stack_ptr(void);
 
-STATIC_INLINE jl_gc_pagemeta_t *page_metadata(void *_data)
+STATIC_INLINE jl_gc_pagemeta_t *page_metadata(void *_data) JL_NOTSAFEPOINT
 {
     uintptr_t data = ((uintptr_t)_data);
     unsigned i;
@@ -305,7 +466,7 @@ struct jl_gc_metadata_ext {
     unsigned pagetable0_i32, pagetable0_i;
 };
 
-STATIC_INLINE struct jl_gc_metadata_ext page_metadata_ext(void *_data)
+STATIC_INLINE struct jl_gc_metadata_ext page_metadata_ext(void *_data) JL_NOTSAFEPOINT
 {
     uintptr_t data = (uintptr_t)_data;
     struct jl_gc_metadata_ext info;
@@ -326,7 +487,7 @@ STATIC_INLINE struct jl_gc_metadata_ext page_metadata_ext(void *_data)
     return info;
 }
 
-STATIC_INLINE void gc_big_object_unlink(const bigval_t *hdr)
+STATIC_INLINE void gc_big_object_unlink(const bigval_t *hdr) JL_NOTSAFEPOINT
 {
     *hdr->prev = hdr->next;
     if (hdr->next) {
@@ -334,7 +495,7 @@ STATIC_INLINE void gc_big_object_unlink(const bigval_t *hdr)
     }
 }
 
-STATIC_INLINE void gc_big_object_link(bigval_t *hdr, bigval_t **list)
+STATIC_INLINE void gc_big_object_link(bigval_t *hdr, bigval_t **list) JL_NOTSAFEPOINT
 {
     hdr->next = *list;
     hdr->prev = list;
@@ -343,17 +504,28 @@ STATIC_INLINE void gc_big_object_link(bigval_t *hdr, bigval_t **list)
     *list = hdr;
 }
 
-void mark_all_roots(jl_ptls_t ptls);
-void gc_mark_object_list(jl_ptls_t ptls, arraylist_t *list, size_t start);
-void visit_mark_stack(jl_ptls_t ptls);
+STATIC_INLINE void gc_mark_sp_init(jl_gc_mark_cache_t *gc_cache, jl_gc_mark_sp_t *sp)
+{
+    sp->pc = gc_cache->pc_stack;
+    sp->data = gc_cache->data_stack;
+    sp->pc_start = gc_cache->pc_stack;
+    sp->pc_end = gc_cache->pc_stack_end;
+}
+
+void gc_mark_queue_all_roots(jl_ptls_t ptls, jl_gc_mark_sp_t *sp);
+void gc_mark_queue_finlist(jl_gc_mark_cache_t *gc_cache, jl_gc_mark_sp_t *sp,
+                           arraylist_t *list, size_t start);
+void gc_mark_loop(jl_ptls_t ptls, jl_gc_mark_sp_t sp);
+void sweep_stack_pools(void);
 void gc_debug_init(void);
-void jl_mark_box_caches(jl_ptls_t ptls);
+
+extern void *gc_mark_label_addrs[_GC_MARK_L_MAX];
 
 // GC pages
 
 void jl_gc_init_page(void);
-NOINLINE jl_gc_pagemeta_t *jl_gc_alloc_page(void);
-void jl_gc_free_page(void *p);
+NOINLINE jl_gc_pagemeta_t *jl_gc_alloc_page(void) JL_NOTSAFEPOINT;
+void jl_gc_free_page(void *p) JL_NOTSAFEPOINT;
 
 // GC debug
 
@@ -374,17 +546,18 @@ void gc_final_pause_end(int64_t t0, int64_t tend);
 #endif
 
 #ifdef GC_TIME
-void gc_time_pool_start(void);
-void gc_time_count_page(int freedall, int pg_skpd);
-void gc_time_pool_end(int sweep_full);
+void gc_time_pool_start(void) JL_NOTSAFEPOINT;
+void gc_time_count_page(int freedall, int pg_skpd) JL_NOTSAFEPOINT;
+void gc_time_pool_end(int sweep_full) JL_NOTSAFEPOINT;
+void gc_time_sysimg_end(uint64_t t0) JL_NOTSAFEPOINT;
 
-void gc_time_big_start(void);
-void gc_time_count_big(int old_bits, int bits);
-void gc_time_big_end(void);
+void gc_time_big_start(void) JL_NOTSAFEPOINT;
+void gc_time_count_big(int old_bits, int bits) JL_NOTSAFEPOINT;
+void gc_time_big_end(void) JL_NOTSAFEPOINT;
 
-void gc_time_mallocd_array_start(void);
-void gc_time_count_mallocd_array(int bits);
-void gc_time_mallocd_array_end(void);
+void gc_time_mallocd_array_start(void) JL_NOTSAFEPOINT;
+void gc_time_count_mallocd_array(int bits) JL_NOTSAFEPOINT;
+void gc_time_mallocd_array_end(void) JL_NOTSAFEPOINT;
 
 void gc_time_mark_pause(int64_t t0, int64_t scanned_bytes,
                         int64_t perm_scanned_bytes);
@@ -393,21 +566,22 @@ void gc_time_sweep_pause(uint64_t gc_end_t, int64_t actual_allocd,
                          int sweep_full);
 #else
 #define gc_time_pool_start()
-STATIC_INLINE void gc_time_count_page(int freedall, int pg_skpd)
+STATIC_INLINE void gc_time_count_page(int freedall, int pg_skpd) JL_NOTSAFEPOINT
 {
     (void)freedall;
     (void)pg_skpd;
 }
-#define gc_time_pool_end(sweep_full)
+#define gc_time_pool_end(sweep_full) (void)(sweep_full)
+#define gc_time_sysimg_end(t0) (void)(t0)
 #define gc_time_big_start()
-STATIC_INLINE void gc_time_count_big(int old_bits, int bits)
+STATIC_INLINE void gc_time_count_big(int old_bits, int bits) JL_NOTSAFEPOINT
 {
     (void)old_bits;
     (void)bits;
 }
 #define gc_time_big_end()
 #define gc_time_mallocd_array_start()
-STATIC_INLINE void gc_time_count_mallocd_array(int bits)
+STATIC_INLINE void gc_time_count_mallocd_array(int bits) JL_NOTSAFEPOINT
 {
     (void)bits;
 }
@@ -461,10 +635,13 @@ extern int gc_verifying;
 #else
 #define gc_verify(ptls)
 #define verify_val(v)
-#define verify_parent1(ty,obj,slot,arg1)
-#define verify_parent2(ty,obj,slot,arg1,arg2)
+#define verify_parent1(ty,obj,slot,arg1) do {} while (0)
+#define verify_parent2(ty,obj,slot,arg1,arg2) do {} while (0)
 #define gc_verifying (0)
 #endif
+int gc_slot_to_fieldidx(void *_obj, void *slot);
+int gc_slot_to_arrayidx(void *_obj, void *begin);
+NOINLINE void gc_mark_loop_unwind(jl_ptls_t ptls, jl_gc_mark_sp_t sp, int pc_offset);
 
 #ifdef GC_DEBUG_ENV
 JL_DLLEXPORT extern jl_gc_debug_env_t jl_gc_debug_env;
@@ -472,7 +649,7 @@ JL_DLLEXPORT extern jl_gc_debug_env_t jl_gc_debug_env;
 int gc_debug_check_other(void);
 int gc_debug_check_pool(void);
 void gc_debug_print(void);
-void gc_scrub_record_task(jl_task_t *ta);
+void gc_scrub_record_task(jl_task_t *ta) JL_NOTSAFEPOINT;
 void gc_scrub(void);
 #else
 #define gc_sweep_always_full 0
@@ -487,7 +664,7 @@ static inline int gc_debug_check_pool(void)
 static inline void gc_debug_print(void)
 {
 }
-static inline void gc_scrub_record_task(jl_task_t *ta)
+static inline void gc_scrub_record_task(jl_task_t *ta) JL_NOTSAFEPOINT
 {
     (void)ta;
 }
@@ -497,11 +674,11 @@ static inline void gc_scrub(void)
 #endif
 
 #ifdef OBJPROFILE
-void objprofile_count(void *ty, int old, int sz);
+void objprofile_count(void *ty, int old, int sz) JL_NOTSAFEPOINT;
 void objprofile_printall(void);
 void objprofile_reset(void);
 #else
-static inline void objprofile_count(void *ty, int old, int sz)
+static inline void objprofile_count(void *ty, int old, int sz) JL_NOTSAFEPOINT
 {
 }
 
@@ -525,6 +702,8 @@ void gc_stats_big_obj(void);
 // For debugging
 void gc_count_pool(void);
 
+size_t jl_array_nbytes(jl_array_t *a) JL_NOTSAFEPOINT;
+
 #ifdef __cplusplus
 }
 #endif
