@@ -7,7 +7,6 @@
 
 #include "llvm-version.h"
 #include "support/dtypes.h"
-#include <sstream>
 
 #include <llvm-c/Core.h>
 #include <llvm-c/Types.h>
@@ -21,10 +20,8 @@
 #include <llvm/IR/LLVMContext.h>
 #include <llvm/IR/MDBuilder.h>
 
-#if defined(JULIA_ENABLE_THREADING)
-#  include <llvm/IR/InlineAsm.h>
-#  include <llvm/Transforms/Utils/BasicBlockUtils.h>
-#endif
+#include <llvm/IR/InlineAsm.h>
+#include <llvm/Transforms/Utils/BasicBlockUtils.h>
 
 #include "julia.h"
 #include "julia_internal.h"
@@ -33,6 +30,8 @@
 
 using namespace llvm;
 
+typedef Instruction TerminatorInst;
+
 std::pair<MDNode*,MDNode*> tbaa_make_child(const char *name, MDNode *parent=nullptr,
                                            bool isConstant=false);
 
@@ -57,21 +56,16 @@ private:
     Type *T_int8;
     Type *T_size;
     PointerType *T_pint8;
-#ifdef JULIA_ENABLE_THREADING
     GlobalVariable *ptls_slot{nullptr};
     GlobalVariable *ptls_offset{nullptr};
     void set_ptls_attrs(CallInst *ptlsStates) const;
     Instruction *emit_ptls_tp(Value *offset, Instruction *insertBefore) const;
     template<typename T> T *add_comdat(T *G) const;
     GlobalVariable *create_aliased_global(Type *T, StringRef name) const;
-#else
-    GlobalVariable *static_tls;
-#endif
     void fix_ptls_use(CallInst *ptlsStates);
     bool runOnModule(Module &M) override;
 };
 
-#ifdef JULIA_ENABLE_THREADING
 void LowerPTLS::set_ptls_attrs(CallInst *ptlsStates) const
 {
     ptlsStates->addAttribute(AttributeList::FunctionIndex, Attribute::ReadNone);
@@ -86,13 +80,13 @@ Instruction *LowerPTLS::emit_ptls_tp(Value *offset, Instruction *insertBefore) c
         // (and therefore the optimization opportunity) from LLVM.
         // Ref https://github.com/JuliaLang/julia/issues/17288
         static const std::string const_asm_str = [&] () {
-            std::stringstream stm;
+            std::string stm;
 #  if defined(_CPU_X86_64_)
-            stm << "movq %fs:0, $0;\naddq $$" << jl_tls_offset << ", $0";
+            raw_string_ostream(stm) << "movq %fs:0, $0;\naddq $$" << jl_tls_offset << ", $0";
 #  else
-            stm << "movl %gs:0, $0;\naddl $$" << jl_tls_offset << ", $0";
+            raw_string_ostream(stm) << "movl %gs:0, $0;\naddl $$" << jl_tls_offset << ", $0";
 #  endif
-            return stm.str();
+            return stm;
         }();
 #  if defined(_CPU_X86_64_)
         const char *dyn_asm_str = "movq %fs:0, $0;\naddq $1, $0";
@@ -178,7 +172,6 @@ inline T *LowerPTLS::add_comdat(T *G) const
 #endif
     return G;
 }
-#endif
 
 void LowerPTLS::fix_ptls_use(CallInst *ptlsStates)
 {
@@ -187,7 +180,6 @@ void LowerPTLS::fix_ptls_use(CallInst *ptlsStates)
         return;
     }
 
-#ifdef JULIA_ENABLE_THREADING
     if (imaging_mode) {
         if (jl_tls_elf_support) {
             // if (offset != 0)
@@ -213,7 +205,7 @@ void LowerPTLS::fix_ptls_use(CallInst *ptlsStates)
             auto getter = new LoadInst(T_ptls_getter, ptls_slot, "", false, ptlsStates);
             getter->setMetadata(llvm::LLVMContext::MD_tbaa, tbaa_const);
             getter->setMetadata(llvm::LLVMContext::MD_invariant_load, MDNode::get(*ctx, None));
-            ptlsStates->setCalledFunction(getter);
+            ptlsStates->setCalledFunction(ptlsStates->getFunctionType(), getter);
             set_ptls_attrs(ptlsStates);
 
             phi->addIncoming(fastTLS, fastTLS->getParent());
@@ -228,7 +220,7 @@ void LowerPTLS::fix_ptls_use(CallInst *ptlsStates)
         auto getter = new LoadInst(T_ptls_getter, ptls_slot, "", false, ptlsStates);
         getter->setMetadata(llvm::LLVMContext::MD_tbaa, tbaa_const);
         getter->setMetadata(llvm::LLVMContext::MD_invariant_load, MDNode::get(*ctx, None));
-        ptlsStates->setCalledFunction(getter);
+        ptlsStates->setCalledFunction(ptlsStates->getFunctionType(), getter);
         set_ptls_attrs(ptlsStates);
     }
     else if (jl_tls_offset != -1) {
@@ -238,13 +230,9 @@ void LowerPTLS::fix_ptls_use(CallInst *ptlsStates)
     else {
         // use the address of the actual getter function directly
         auto val = ConstantInt::get(T_size, (uintptr_t)jl_get_ptls_states_getter());
-        ptlsStates->setCalledFunction(ConstantExpr::getIntToPtr(val, T_ptls_getter));
+        ptlsStates->setCalledFunction(ptlsStates->getFunctionType(), ConstantExpr::getIntToPtr(val, T_ptls_getter));
         set_ptls_attrs(ptlsStates);
     }
-#else
-    ptlsStates->replaceAllUsesWith(static_tls);
-    ptlsStates->eraseFromParent();
-#endif
 }
 
 bool LowerPTLS::runOnModule(Module &_M)
@@ -264,20 +252,15 @@ bool LowerPTLS::runOnModule(Module &_M)
     T_int8 = Type::getInt8Ty(*ctx);
     T_size = sizeof(size_t) == 8 ? Type::getInt64Ty(*ctx) : Type::getInt32Ty(*ctx);
     T_pint8 = T_int8->getPointerTo();
-#ifdef JULIA_ENABLE_THREADING
     if (imaging_mode) {
         ptls_slot = create_aliased_global(T_ptls_getter, "jl_get_ptls_states_slot");
         ptls_offset = create_aliased_global(T_size, "jl_tls_offset");
     }
-#else
-    static_tls = new GlobalVariable(*M, T_ppjlvalue, false, GlobalVariable::ExternalLinkage,
-                                    NULL, "jl_tls_states");
-#endif
 
     for (auto it = ptls_getter->user_begin(); it != ptls_getter->user_end();) {
         auto call = cast<CallInst>(*it);
         ++it;
-        assert(call->getCalledValue() == ptls_getter);
+        assert(call->getCalledOperand() == ptls_getter);
         fix_ptls_use(call);
     }
     assert(ptls_getter->use_empty());
