@@ -19,7 +19,7 @@ extern "C" {
 // 1: at least one sigint is pending, only the sigint page is enabled.
 // 2: at least one sigint is pending, both safepoint pages are enabled.
 JL_DLLEXPORT sig_atomic_t jl_signal_pending = 0;
-volatile uint32_t jl_gc_running = 0;
+uint32_t jl_gc_running = 0;
 char *jl_safepoint_pages = NULL;
 // The number of safepoints enabled on the three pages.
 // The first page, is the SIGINT page, only used by the master thread.
@@ -157,9 +157,9 @@ void jl_safepoint_wait_gc(void)
 {
     // The thread should have set this is already
     assert(jl_get_ptls_states()->gc_state != 0);
-    // Use normal volatile load in the loop.
-    // Use a acquire load to make sure the GC result is visible on this thread.
-    while (jl_gc_running || jl_atomic_load_acquire(&jl_gc_running)) {
+    // Use normal volatile load in the loop for speed until GC finishes.
+    // Then use an acquire load to make sure the GC result is visible on this thread.
+    while (jl_atomic_load_relaxed(&jl_gc_running) || jl_atomic_load_acquire(&jl_gc_running)) {
         jl_cpu_pause(); // yield?
     }
 }
