@@ -1,5 +1,7 @@
 // This file is a part of Julia. License is MIT: https://julialang.org/license
 
+// Note that this file is `#include`d by "signals-unix.c"
+
 #include <mach/clock.h>
 #include <mach/clock_types.h>
 #include <mach/clock_reply.h>
@@ -17,6 +19,18 @@
 
 #include "julia_assert.h"
 
+// private keymgr stuff
+#define KEYMGR_GCC3_DW2_OBJ_LIST 302
+enum {
+  NM_ALLOW_RECURSION = 1,
+  NM_RECURSION_ILLEGAL = 2
+};
+extern void _keymgr_set_and_unlock_processwide_ptr(unsigned int key, void *ptr);
+extern int _keymgr_unlock_processwide_ptr(unsigned int key);
+extern void *_keymgr_get_and_lock_processwide_ptr(unsigned int key);
+extern int _keymgr_get_and_lock_processwide_ptr_2(unsigned int key, void **result);
+extern int _keymgr_set_lockmode_processwide_ptr(unsigned int key, unsigned int mode);
+
 static void attach_exception_port(thread_port_t thread, int segv_only);
 
 // low 16 bits are the thread id, the next 8 bits are the original gc_state
@@ -24,13 +38,13 @@ static arraylist_t suspended_threads;
 void jl_mach_gc_end(void)
 {
     // Requires the safepoint lock to be held
-    for (size_t i = 0;i < suspended_threads.len;i++) {
+    for (size_t i = 0; i < suspended_threads.len; i++) {
         uintptr_t item = (uintptr_t)suspended_threads.items[i];
         int16_t tid = (int16_t)item;
         int8_t gc_state = (int8_t)(item >> 8);
         jl_ptls_t ptls2 = jl_all_tls_states[tid];
         jl_atomic_store_release(&ptls2->gc_state, gc_state);
-        thread_resume(pthread_mach_thread_np((pthread_t)ptls2->system_id));
+        thread_resume(pthread_mach_thread_np(ptls2->system_id));
     }
     suspended_threads.len = 0;
 }
@@ -41,9 +55,10 @@ static int jl_mach_gc_wait(jl_ptls_t ptls2,
                            mach_port_t thread, int16_t tid)
 {
     jl_mutex_lock_nogc(&safepoint_lock);
-    if (!jl_gc_running) {
-        // GC is done before we get the message or the safepoint is enabled
-        // for SIGINT.
+    if (!jl_atomic_load_relaxed(&jl_gc_running)) {
+        // relaxed, since gets set to zero only while the safepoint_lock was held
+        // this means we can tell if GC is done before we got the message or
+        // the safepoint was enabled for SIGINT.
         jl_mutex_unlock_nogc(&safepoint_lock);
         return 0;
     }
@@ -78,6 +93,17 @@ void *mach_segv_listener(void *arg)
 
 static void allocate_segv_handler()
 {
+    // ensure KEYMGR_GCC3_DW2_OBJ_LIST is initialized, as this requires malloc
+    // and thus can deadlock when used without first initializing it.
+    // Apple caused this problem in their libunwind in 10.9 (circa keymgr-28)
+    // when they removed this part of the code from keymgr.
+    // Much thanks to Apple for providing source code, or this would probably
+    // have simply remained unsolved forever on their platform.
+    // This is similar to just calling checkKeyMgrRegisteredFDEs
+    // (this is quite thread-unsafe)
+    if (_keymgr_set_lockmode_processwide_ptr(KEYMGR_GCC3_DW2_OBJ_LIST, NM_ALLOW_RECURSION))
+        jl_error("_keymgr_set_lockmode_processwide_ptr failed");
+
     arraylist_new(&suspended_threads, jl_n_threads);
     pthread_t thread;
     pthread_attr_t attr;
@@ -97,7 +123,7 @@ static void allocate_segv_handler()
     }
     pthread_attr_destroy(&attr);
     for (int16_t tid = 0;tid < jl_n_threads;tid++) {
-        attach_exception_port(pthread_mach_thread_np((pthread_t)jl_all_tls_states[tid]->system_id), 0);
+        attach_exception_port(pthread_mach_thread_np(jl_all_tls_states[tid]->system_id), 0);
     }
 }
 
@@ -112,13 +138,30 @@ static kern_return_t profiler_segv_handler
                  mach_msg_type_number_t                   code_count);
 #endif
 
+#if defined(_CPU_X86_64_)
+typedef x86_thread_state64_t host_thread_state_t;
+typedef x86_exception_state64_t host_exception_state_t;
+#define THREAD_STATE x86_THREAD_STATE64
+#define THREAD_STATE_COUNT x86_THREAD_STATE64_COUNT
+#define HOST_EXCEPTION_STATE x86_EXCEPTION_STATE64
+#define HOST_EXCEPTION_STATE_COUNT x86_EXCEPTION_STATE64_COUNT
+
 enum x86_trap_flags {
     USER_MODE = 0x4,
     WRITE_FAULT = 0x2,
     PAGE_PRESENT = 0x1
 };
 
-static void jl_call_in_state(jl_ptls_t ptls2, x86_thread_state64_t *state,
+#elif defined(_CPU_AARCH64_)
+typedef arm_thread_state64_t host_thread_state_t;
+typedef arm_exception_state64_t host_exception_state_t;
+#define THREAD_STATE ARM_THREAD_STATE64
+#define THREAD_STATE_COUNT ARM_THREAD_STATE64_COUNT
+#define HOST_EXCEPTION_STATE ARM_EXCEPTION_STATE64
+#define HOST_EXCEPTION_STATE_COUNT ARM_EXCEPTION_STATE64_COUNT
+#endif
+
+static void jl_call_in_state(jl_ptls_t ptls2, host_thread_state_t *state,
                              void (*fptr)(void))
 {
     uint64_t rsp = (uint64_t)ptls2->signal_stack + sig_stack_size;
@@ -128,25 +171,30 @@ static void jl_call_in_state(jl_ptls_t ptls2, x86_thread_state64_t *state,
     rsp -= sizeof(void*);
     *(void**)rsp = NULL;
 
+#ifdef _CPU_X86_64_
     state->__rsp = rsp; // set stack pointer
     state->__rip = (uint64_t)fptr; // "call" the function
+#else
+    state->__sp = rsp;
+    state->__pc = (uint64_t)fptr;
+#endif
 }
 
 static void jl_throw_in_thread(int tid, mach_port_t thread, jl_value_t *exception)
 {
-    unsigned int count = MACHINE_THREAD_STATE_COUNT;
-    x86_thread_state64_t state;
-    kern_return_t ret = thread_get_state(thread, x86_THREAD_STATE64, (thread_state_t)&state, &count);
+    unsigned int count = THREAD_STATE_COUNT;
+    host_thread_state_t state;
+    kern_return_t ret = thread_get_state(thread, THREAD_STATE, (thread_state_t)&state, &count);
     HANDLE_MACH_ERROR("thread_get_state", ret);
     jl_ptls_t ptls2 = jl_all_tls_states[tid];
     if (!ptls2->safe_restore) {
         assert(exception);
         ptls2->bt_size = rec_backtrace_ctx(ptls2->bt_data, JL_MAX_BT_SIZE,
-                                           (bt_context_t*)&state, ptls2->pgcstack, 0);
+                                           (bt_context_t*)&state, ptls2->pgcstack);
         ptls2->sig_exception = exception;
     }
     jl_call_in_state(ptls2, &state, &jl_sig_throw);
-    ret = thread_set_state(thread, x86_THREAD_STATE64,
+    ret = thread_set_state(thread, THREAD_STATE,
                            (thread_state_t)&state, count);
     HANDLE_MACH_ERROR("thread_set_state", ret);
 }
@@ -160,10 +208,10 @@ kern_return_t catch_exception_raise(mach_port_t            exception_port,
                                     exception_data_t       code,
                                     mach_msg_type_number_t code_count)
 {
-    unsigned int count = MACHINE_THREAD_STATE_COUNT;
-    unsigned int exc_count = X86_EXCEPTION_STATE64_COUNT;
-    x86_exception_state64_t exc_state;
-    x86_thread_state64_t state;
+    unsigned int count = THREAD_STATE_COUNT;
+    unsigned int exc_count = HOST_EXCEPTION_STATE_COUNT;
+    host_exception_state_t exc_state;
+    host_thread_state_t state;
 #ifdef LIBOSXUNWIND
     if (thread == mach_profiler_thread) {
         return profiler_segv_handler(exception_port, thread, task, exception, code, code_count);
@@ -173,7 +221,7 @@ kern_return_t catch_exception_raise(mach_port_t            exception_port,
     jl_ptls_t ptls2 = NULL;
     for (tid = 0;tid < jl_n_threads;tid++) {
         jl_ptls_t _ptls2 = jl_all_tls_states[tid];
-        if (pthread_mach_thread_np((pthread_t)_ptls2->system_id) == thread) {
+        if (pthread_mach_thread_np(_ptls2->system_id) == thread) {
             ptls2 = _ptls2;
             break;
         }
@@ -190,9 +238,13 @@ kern_return_t catch_exception_raise(mach_port_t            exception_port,
         return KERN_SUCCESS;
     }
     assert(exception == EXC_BAD_ACCESS);
-    kern_return_t ret = thread_get_state(thread, x86_EXCEPTION_STATE64, (thread_state_t)&exc_state, &exc_count);
+    kern_return_t ret = thread_get_state(thread, HOST_EXCEPTION_STATE, (thread_state_t)&exc_state, &exc_count);
     HANDLE_MACH_ERROR("thread_get_state", ret);
+#ifdef _CPU_X86_64_
     uint64_t fault_addr = exc_state.__faultvaddr;
+#else
+    uint64_t fault_addr = exc_state.__far;
+#endif
     if (jl_addr_is_safepoint(fault_addr)) {
         if (jl_mach_gc_wait(ptls2, thread, tid))
             return KERN_SUCCESS;
@@ -236,7 +288,7 @@ kern_return_t catch_exception_raise(mach_port_t            exception_port,
         return KERN_SUCCESS;
     }
     else {
-        kern_return_t ret = thread_get_state(thread, x86_THREAD_STATE64, (thread_state_t)&state, &count);
+        kern_return_t ret = thread_get_state(thread, THREAD_STATE, (thread_state_t)&state, &count);
         HANDLE_MACH_ERROR("thread_get_state", ret);
         jl_critical_error(SIGSEGV, (unw_context_t*)&state,
                           ptls2->bt_data, &ptls2->bt_size);
@@ -258,18 +310,18 @@ static void attach_exception_port(thread_port_t thread, int segv_only)
 static void jl_thread_suspend_and_get_state(int tid, unw_context_t **ctx)
 {
     jl_ptls_t ptls2 = jl_all_tls_states[tid];
-    mach_port_t tid_port = pthread_mach_thread_np((pthread_t)ptls2->system_id);
+    mach_port_t tid_port = pthread_mach_thread_np(ptls2->system_id);
 
     kern_return_t ret = thread_suspend(tid_port);
     HANDLE_MACH_ERROR("thread_suspend", ret);
 
     // Do the actual sampling
-    unsigned int count = MACHINE_THREAD_STATE_COUNT;
+    unsigned int count = THREAD_STATE_COUNT;
     static unw_context_t state;
     memset(&state, 0, sizeof(unw_context_t));
 
     // Get the state of the suspended thread
-    ret = thread_get_state(tid_port, x86_THREAD_STATE64, (thread_state_t)&state, &count);
+    ret = thread_get_state(tid_port, THREAD_STATE, (thread_state_t)&state, &count);
 
     // Initialize the unwind context with the suspend thread's state
     *ctx = &state;
@@ -278,7 +330,7 @@ static void jl_thread_suspend_and_get_state(int tid, unw_context_t **ctx)
 static void jl_thread_resume(int tid, int sig)
 {
     jl_ptls_t ptls2 = jl_all_tls_states[tid];
-    mach_port_t thread = pthread_mach_thread_np((pthread_t)ptls2->system_id);
+    mach_port_t thread = pthread_mach_thread_np(ptls2->system_id);
     kern_return_t ret = thread_resume(thread);
     HANDLE_MACH_ERROR("thread_resume", ret);
 }
@@ -288,7 +340,7 @@ static void jl_thread_resume(int tid, int sig)
 static void jl_try_deliver_sigint(void)
 {
     jl_ptls_t ptls2 = jl_all_tls_states[0];
-    mach_port_t thread = pthread_mach_thread_np((pthread_t)ptls2->system_id);
+    mach_port_t thread = pthread_mach_thread_np(ptls2->system_id);
 
     kern_return_t ret = thread_suspend(thread);
     HANDLE_MACH_ERROR("thread_suspend", ret);
@@ -317,7 +369,7 @@ static void jl_try_deliver_sigint(void)
 static void jl_exit_thread0(int exitstate)
 {
     jl_ptls_t ptls2 = jl_all_tls_states[0];
-    mach_port_t thread = pthread_mach_thread_np((pthread_t)ptls2->system_id);
+    mach_port_t thread = pthread_mach_thread_np(ptls2->system_id);
     kern_return_t ret = thread_suspend(thread);
     HANDLE_MACH_ERROR("thread_suspend", ret);
 
@@ -325,9 +377,9 @@ static void jl_exit_thread0(int exitstate)
     ret = thread_abort(thread);
     HANDLE_MACH_ERROR("thread_abort", ret);
 
-    unsigned int count = MACHINE_THREAD_STATE_COUNT;
-    x86_thread_state64_t state;
-    ret = thread_get_state(thread, x86_THREAD_STATE64,
+    unsigned int count = THREAD_STATE_COUNT;
+    host_thread_state_t state;
+    ret = thread_get_state(thread, THREAD_STATE,
                            (thread_state_t)&state, &count);
 
     void (*exit_func)(int) = &_exit;
@@ -338,10 +390,16 @@ static void jl_exit_thread0(int exitstate)
         exit_func = &exit;
     }
 
+#ifdef _CPU_X86_64_
     // First integer argument. Not portable but good enough =)
     state.__rdi = exitstate;
+#elif defined(_CPU_AARCH64_)
+    state.__x[0] = exitstate;
+#else
+#error Fill in first integer argument here
+#endif
     jl_call_in_state(ptls2, &state, (void (*)(void))exit_func);
-    ret = thread_set_state(thread, x86_THREAD_STATE64,
+    ret = thread_set_state(thread, THREAD_STATE,
                            (thread_state_t)&state, count);
     HANDLE_MACH_ERROR("thread_set_state", ret);
 
@@ -368,7 +426,7 @@ static kern_return_t profiler_segv_handler
                  mach_msg_type_number_t                   code_count)
 {
     assert(thread == mach_profiler_thread);
-    x86_thread_state64_t state;
+    host_thread_state_t state;
 
     // Not currently unwinding. Raise regular segfault
     if (forceDwarf == -2)
@@ -379,24 +437,34 @@ static kern_return_t profiler_segv_handler
     else
         forceDwarf = -1;
 
-    unsigned int count = MACHINE_THREAD_STATE_COUNT;
+    unsigned int count = THREAD_STATE_COUNT;
 
-    thread_get_state(thread, x86_THREAD_STATE64, (thread_state_t)&state, &count);
+    thread_get_state(thread, THREAD_STATE, (thread_state_t)&state, &count);
 
+#ifdef _CPU_X86_64_
     // don't change cs fs gs rflags
     uint64_t cs = state.__cs;
     uint64_t fs = state.__fs;
     uint64_t gs = state.__gs;
     uint64_t rflags = state.__rflags;
+#elif defined(_CPU_AARCH64_)
+    uint64_t cpsr = state.__cpsr;
+#else
+#error Unknown CPU
+#endif
 
-    memcpy(&state, &profiler_uc, sizeof(x86_thread_state64_t));
+    memcpy(&state, &profiler_uc, sizeof(state));
 
+#ifdef _CPU_X86_64_
     state.__cs = cs;
     state.__fs = fs;
     state.__gs = gs;
     state.__rflags = rflags;
+#else
+    state.__cpsr = cpsr;
+#endif
 
-    kern_return_t ret = thread_set_state(thread, x86_THREAD_STATE64, (thread_state_t)&state, count);
+    kern_return_t ret = thread_set_state(thread, THREAD_STATE, (thread_state_t)&state, count);
     HANDLE_MACH_ERROR("thread_set_state", ret);
 
     return KERN_SUCCESS;
@@ -420,10 +488,15 @@ void *mach_profile_listener(void *arg)
         HANDLE_MACH_ERROR("mach_msg", ret);
         // sample each thread, round-robin style in reverse order
         // (so that thread zero gets notified last)
+        jl_lock_profile();
+        void *unused = NULL;
+        int keymgr_locked = _keymgr_get_and_lock_processwide_ptr_2(KEYMGR_GCC3_DW2_OBJ_LIST, &unused) == 0;
         for (i = jl_n_threads; i-- > 0; ) {
             // if there is no space left, break early
-            if (bt_size_cur >= bt_size_max - 1)
+            if (jl_profile_is_buffer_full()) {
+                jl_profile_stop_timer();
                 break;
+            }
 
             unw_context_t *uc;
             jl_thread_suspend_and_get_state(i, &uc);
@@ -448,7 +521,7 @@ void *mach_profile_listener(void *arg)
 
                 if (forceDwarf == 0) {
                     // Save the backtrace
-                    bt_size_cur += rec_backtrace_ctx((jl_bt_element_t*)bt_data_prof + bt_size_cur, bt_size_max - bt_size_cur - 1, uc, NULL, 1);
+                    bt_size_cur += rec_backtrace_ctx((jl_bt_element_t*)bt_data_prof + bt_size_cur, bt_size_max - bt_size_cur - 1, uc, NULL);
                 }
                 else if (forceDwarf == 1) {
                     bt_size_cur += rec_backtrace_ctx_dwarf((jl_bt_element_t*)bt_data_prof + bt_size_cur, bt_size_max - bt_size_cur - 1, uc, NULL);
@@ -464,14 +537,18 @@ void *mach_profile_listener(void *arg)
 
                 // Mark the end of this block with 0
                 bt_data_prof[bt_size_cur++].uintptr = 0;
-
-                // Reset the alarm
-                kern_return_t ret = clock_alarm(clk, TIME_RELATIVE, timerprof, profile_port);
-                HANDLE_MACH_ERROR("clock_alarm", ret)
             }
             // We're done! Resume the thread.
             jl_thread_resume(i, 0);
         }
+        if (keymgr_locked)
+            _keymgr_unlock_processwide_ptr(KEYMGR_GCC3_DW2_OBJ_LIST);
+        jl_unlock_profile();
+        if (running) {
+            // Reset the alarm
+            kern_return_t ret = clock_alarm(clk, TIME_RELATIVE, timerprof, profile_port);
+            HANDLE_MACH_ERROR("clock_alarm", ret)
+        }
     }
 }
 
@@ -505,6 +582,7 @@ JL_DLLEXPORT int jl_profile_start_timer(void)
     timerprof.tv_nsec = nsecprof%GIGA;
 
     running = 1;
+    // ensure the alarm is running
     ret = clock_alarm(clk, TIME_RELATIVE, timerprof, profile_port);
     HANDLE_MACH_ERROR("clock_alarm", ret);
 
